# To install:
# python3 -m pip install -r requirements.txt
import duckdb
import geopandas as gpd
import json
import matplotlib
import matplotlib.pyplot as plt
import numpy
import pandas
import seaborn as sns
import scipy.optimize as optim
from sklearn.ensemble import HistGradientBoostingRegressor


def is_optimal_gain(df) -> float:
    """
    Computes the maximum potential efficiency gain.
    """
    size = df.fire_size
    cost = df.fire_cost
    d = df.mu_over_nu # INTRO mu_over_nu denotes the sub-sampling weight.
    return (d*size).sum()*(d*cost).sum()*((d*((size*cost)**0.5)).sum()**-2)


def is_optimal_gain_bstd(df, n_bootstrap_replicas=100) -> float:
    """
    Bootstrap estimate of standard error for the maximum potential efficiency gain.
    """
    def draw_bootstrap_replica():
        return df.sample(frac=1, replace=True)
    return numpy.std(numpy.array([is_optimal_gain(draw_bootstrap_replica()) for _i in range(n_bootstrap_replicas)]))


def norm_scalar(df: pandas.DataFrame, dp_dq: pandas.Series) -> float:
    """
    Computes the normalization constant by which to multiply weights `w`
    in order to get a sampling distribution with the same expected cost as in `df`.
    """
    c = df.fire_cost
    d = df.mu_over_nu
    ref_cost = (d*c).sum()
    c_q = (d*c/dp_dq).sum()
    norm_scalr = c_q/ref_cost # normalized weights
    return norm_scalr


def score_weight_fn(df, w_fn, params):
    """
    Returns a variance reduction factor (inverse of efficiency gain).
    """
    A = df.fire_size
    d = df.mu_over_nu
    w1 = w_fn(df, *params) # non-normalized weights
    norm_scalr = norm_scalar(df, w1)
    w2 = w1 * norm_scalr # the weight is normalized so that the cost is the same as a no-weigthing situation.
    V = (d*w2*A).sum() # variance estimate
    red = (V / (d*A).sum())
    return red



#### Reweighting pyromes uniformly

def esum_cond_categorical(df, x_col, y_col):
    """
    For each categorical value x in `df[x_col]`,
    computes the expected sum of Y (`y_col`) given X=x.
    """
    df1 = df[['mu_over_nu']].assign(x=df[x_col],y=df[y_col])
    df2 = duckdb.sql("""
        SELECT
            x,
            SUM(mu_over_nu * y) AS e_y
        FROM df1
        GROUP BY x
    """).df()
    return pandas.DataFrame({x_col: df2.x, y_col: df2.e_y})


def pyrome_weights_series(df) -> pandas.Series:
    dfa = esum_cond_categorical(df, 'pyrome_id', 'fire_size')
    dfc = esum_cond_categorical(df, 'pyrome_id', 'fire_cost')
    df1 = esum_cond_categorical(df.assign(n_fires=lambda _f: 1), 'pyrome_id', 'n_fires')
    dfw = pandas.merge(dfa, dfc, how='inner', on='pyrome_id').merge(df1, how='inner', on='pyrome_id').assign(
        w0 = lambda p: (p.fire_cost/p.fire_size)**0.5
        )
    df2 = pandas.merge(df, dfw[['pyrome_id', 'w0']], how='left', on='pyrome_id') # Left join is important to preserve index/order
    return df2.w0 * norm_scalar(df2, df2.w0)


def df_pyrome_multipliers(df):
    df1 = df[lambda f: f.pyrome_id > 0].assign(pyr_mul=1.0/pyrome_weights_series(df))
    return duckdb.sql("""
        SELECT
            pyrome_id, pyr_mul,
            SUM(mu_over_nu * fire_a_total_ac) AS esum_burned_ac,
            SUM(mu_over_nu * fire_runtime) AS esum_runtime_s,
            SUM(mu_over_nu) AS esum_1
        FROM df1 GROUP BY ALL
        ORDER BY pyrome_id""").df()


#### Fitting a reweighting model

def fit_reweighting_model(df, w_fn, init_params, minimize_kwargs={}):
    optres = optim.minimize(
        lambda p: score_weight_fn(df, w_fn, p),
        numpy.array(init_params),
        **minimize_kwargs)
    return {
        'efficiency_gain': 1.0/optres.fun,
        'bestfit_params': optres.x,
        'minimize_result': optres
        } 


#### w1 reweighting model

def w1(f, alpha):
    return numpy.exp(alpha * numpy.exp(- f.fire_tstop_hr/150.0))




#### w2 reweighting model

def w2(f, alpha1, t1, alpha2, t2):
    return numpy.exp(alpha1 * numpy.exp(- f.fire_tstop_hr/t1) + alpha2 * numpy.exp(- f.fire_tstop_hr/t2))


#### w3: gradient-boosted regression tree

def df_fuel_models():
    """
    A DataFrame of 'fire aggressiveness' metric computed for each fuel models.
    These numbers were computed externally using Rothermel equations,
    with wind speed of 10mi/hr, no slope, and reference fuel moistures.
    """
    df_fm = pandas.DataFrame({'fbfm40': [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 91, 92, 93, 98, 99, 101, 102, 103, 104, 105, 106, 107, 108, 109, 121, 122, 123, 124, 141, 142, 143, 144, 145, 146, 147, 148, 149, 161, 162, 163, 164, 165, 181, 182, 183, 184, 185, 186, 187, 188, 189, 201, 202, 203, 204, 211, 212, 213, 214, 215, 216, 217, 218, 221, 222, 223, 224, 241, 242, 243, 244, 245, 246, 247, 248, 249, 261, 262, 263, 265, 281, 282, 283, 284, 285, 286, 287, 288, 289, 301, 302, 303], 'ros_ws10mph_fpm': [491.10546107832647, 203.48047900893104, 425.79921291235337, 456.3585056950752, 153.89739519779602, 123.3292676344611, 128.75214011972344, 7.71827644400651, 40.14225564941187, 46.34625593731968, 18.40971278226012, 40.51722432534907, 48.978039048884625, 0.0, 0.0, 0.0, 0.0, 0.0, 14.79737064432105, 87.92117106667558, 103.90455140433436, 177.27325556315452, 136.59537427832313, 197.95784828850992, 258.37699232225015, 227.03488586539862, 401.32902353976806, 64.71486378087454, 89.91986930610565, 130.00347383765885, 106.48484504422197, 3.7788924959316996, 24.397274864463704, 12.719115867792906, 107.08303599924203, 177.81511077162196, 70.11599683523295, 114.34864933021883, 68.46624154953042, 126.89879887816151, 10.3330588348844, 44.95206672554619, 99.11978122136715, 49.942227989858935, 27.4432030244224, 1.5916422420211531, 4.117332340153939, 5.703647377937049, 8.730743541576157, 16.65038420232926, 23.63826420467111, 8.87759125497545, 22.139347184960293, 32.75982876774987, 22.931331375546876, 59.91086228506879, 110.51546503449502, 213.1540890803797, 14.79737064432105, 87.92117106667558, 103.90455140433436, 177.27325556315452, 136.59537427832313, 197.95784828850992, 258.37699232225015, 227.03488586539862, 64.71486378087454, 89.91986930610565, 130.00347383765885, 106.48484504422197, 3.7788924959316996, 24.397274864463704, 12.719115867792906, 107.08303599924203, 177.81511077162196, 70.11599683523295, 114.34864933021883, 68.46624154953042, 126.89879887816151, 10.3330588348844, 44.95206672554619, 99.11978122136715, 27.4432030244224, 1.5916422420211531, 4.117332340153939, 5.703647377937049, 8.730743541576157, 16.65038420232926, 23.63826420467111, 8.87759125497545, 22.139347184960293, 32.75982876774987, 22.931331375546876, 59.91086228506879, 110.51546503449502], 'rx_intensity_kW_m2': [177.80398597703808, 770.0886179865292, 665.5646928112095, 2875.2467062008786, 678.8669741953732, 437.59174699200724, 465.2841489843138, 209.85809884124896, 549.0617655950334, 1376.179788245684, 513.2196335947764, 1488.733473628445, 2225.3297240467127, 0.0, 0.0, 0.0, 0.0, 0.0, 92.05810372438549, 244.20644224094642, 308.35500875390966, 474.61445065308857, 865.2847251746248, 1325.3132841854215, 1666.6406919956714, 2010.1090619820507, 3042.0773009564323, 291.7963094768179, 476.37700543086584, 772.0560890756184, 3378.7099467876023, 89.03640809216672, 1164.84263225473, 389.0404326910691, 686.5529867569288, 1118.7901346194417, 1052.2342695063694, 1485.4257820320163, 1615.1010057465483, 2567.833796489643, 350.80865284909385, 408.7681840166796, 793.8916444379249, 1363.3180245951182, 1768.2676138510153, 105.7981562530721, 155.8281907711614, 177.7866526305196, 221.8727946233221, 354.05272076150237, 508.84348085366724, 356.5686416096109, 726.739679537692, 1052.4479635442547, 556.8282234923714, 1127.0066760463862, 1553.747921846095, 1636.8439952093277, 92.05810372438549, 244.20644224094642, 308.35500875390966, 474.61445065308857, 865.2847251746248, 1325.3132841854215, 1666.6406919956714, 2010.1090619820507, 291.7963094768179, 476.37700543086584, 772.0560890756184, 3378.7099467876023, 89.03640809216672, 1164.84263225473, 389.0404326910691, 686.5529867569288, 1118.7901346194417, 1052.2342695063694, 1485.4257820320163, 1615.1010057465483, 2567.833796489643, 350.80865284909385, 408.7681840166796, 793.8916444379249, 1768.2676138510153, 105.7981562530721, 155.8281907711614, 177.7866526305196, 221.8727946233221, 354.05272076150237, 508.84348085366724, 356.5686416096109, 726.739679537692, 1052.4479635442547, 556.8282234923714, 1127.0066760463862, 1553.747921846095], 'res_time_min': [0.10971428571428571, 0.1379302056279426, 0.256, 0.22078636586528197, 0.22818499612124019, 0.2455390273287783, 0.24583849973709135, 0.20329720120828762, 0.15460302097864495, 0.21764611427384198, 0.32495697795882206, 0.3354756859246041, 0.3313244389821007, 'Infinity', 'Infinity', 'Infinity', 'Infinity', 'Infinity', 0.1869833729216152, 0.2109844658799876, 0.2977304652845389, 0.21034627185506258, 0.23550285145288313, 0.19139905731570028, 0.20936412389186823, 0.2949352692969466, 0.2381977178408828, 0.20955665024630543, 0.21012730729142198, 0.23794703036382753, 0.2354222780667724, 0.22935677631524715, 0.22971884333771272, 0.27999537634510213, 0.2283338581461741, 0.30680487360142095, 0.3355682876830045, 0.3114386108657176, 0.276969627885575, 0.278656926629776, 0.2390077658246982, 0.217382350167601, 0.23831709250548191, 0.17325463628182453, 0.3138266174914029, 0.2237534007461343, 0.2125966472313487, 0.2504714934367811, 0.24484799550367067, 0.22412908981846083, 0.19835510292868358, 0.3121571513543511, 0.21692336615184582, 0.22152102232360932, 0.23226397415189165, 0.2038298116632697, 0.19848542294147123, 0.20141321485011784, 0.1869833729216152, 0.2109844658799876, 0.2977304652845389, 0.21034627185506258, 0.23550285145288313, 0.19139905731570028, 0.20936412389186823, 0.2949352692969466, 0.20955665024630543, 0.21012730729142198, 0.23794703036382753, 0.2354222780667724, 0.22935677631524715, 0.22971884333771272, 0.27999537634510213, 0.2283338581461741, 0.30680487360142095, 0.3355682876830045, 0.3114386108657176, 0.276969627885575, 0.278656926629776, 0.2390077658246982, 0.217382350167601, 0.23831709250548191, 0.3138266174914029, 0.2237534007461343, 0.2125966472313487, 0.2504714934367811, 0.24484799550367067, 0.22412908981846083, 0.19835510292868358, 0.3121571513543511, 0.21692336615184582, 0.22152102232360932, 0.23226397415189165, 0.2038298116632697, 0.19848542294147123]})
    df_fm['fli_ws10mph_kW_m'] = df_fm.rx_intensity_kW_m2 * (df_fm.res_time_min.astype(float) * df_fm.ros_ws10mph_fpm * 0.3048)
    return df_fm

w3_all_feature_cols = {
    'pyrome_id',
    'fm_90', # Fuel model index (offset by -90 to appease Scikit-Learn)
    'ign_ercperc_v3', # percentile of ERC(G) - proxy of dryness anomaly.
    'fire_tstop_hr', # Fire duration.
    'ros_ws10mph_fpm', 'rx_intensity_kW_m2', 'fli_ws10mph_kW_m', # Fire aggressiveness metrics of the fuel model.
    'x', 'y' # Spatial coordinates in EPSG:5070
    }
w3_cat_features = {'pyrome_id', 'fm_90'}

def w3_X(df, feature_cols):
    """
    Computes the features matrix of the w3 model.
    """
    feature_cols = list(sorted(feature_cols)) # to make iteration order predictable.
    df = df.assign(fm_90 = lambda f: f.fbfm40_23_11_nbflip - 90).drop(columns=['fbfm40'])
    df = pandas.merge(df, df_fuel_models(), left_on='fbfm40_23_11_nbflip', right_on='fbfm40', how='left')
    return numpy.array([
        df[k].to_numpy() for k in feature_cols
    ], dtype=float).T

def y_w_optim(df):
    """
    Computes the target variable vector of the w3 model.
    """
    return 0.5*numpy.log(df.fire_cost/df.fire_size)


def w3_train_model(df_train, feature_cols):
    feature_cols = list(sorted(feature_cols)) # to make iteration order predictable.
    X_categorical_features = [(c in w3_cat_features) for c in feature_cols]
    est = HistGradientBoostingRegressor(categorical_features=X_categorical_features).fit(
        w3_X(df_train, feature_cols),
        y_w_optim(df_train)
        , sample_weight=(df_train.mu_over_nu * df_train.fire_a_total_ac).to_numpy()
    )
    return est


def w3_test_kfold(df, feature_cols=w3_all_feature_cols, train_frac=0.7, n_iter=30):
    n = len(df)
    n_train = int(round(n*train_frac))
    def evaluate():
        df1 = df.sample(frac=1) # shuffles the rows.
        df_train = df.head(n_train)
        est = w3_train_model(df_train, feature_cols)
        df_test = df.tail(n - n_train)
        w_pred = numpy.exp(est.predict(w3_X(df_test, feature_cols)))
        effic_gain = 1.0/score_weight_fn(df_test.assign(w_pred=w_pred), lambda f: f.w_pred, [])
        return effic_gain
    return [evaluate() for _i in range(n_iter)]


### Data

def read_df_hist_fires(fpath='~/projects/SIG/ffv3_predictors/conus_historical_fires_1992to2020.parquet'):
    """
    Loads and returns a DataFrame of historical fires (enriched version of the FPA-FOD).
    """
    df = pandas.read_parquet(fpath)
    df = df[lambda f: (f.pyrome_id > 0)] # Restricting to CONUS
    df = df.assign(fire_size = lambda f: f.hist_fire_size_ac*0.40468564, mu_over_nu = lambda f: 1.0)
    return df


def read_df_ff_v3_n2e6(fpath='/Users/val/projects/SIG/v3prod5_fires_n2e6.parquet'):
    """
    Reads the pandas DataFrame of (thinned) FireFactor simulations results (about 2.7 million fires).
    """
    df_v3_n2e6 = pandas.read_parquet(fpath).assign(
        fire_size=lambda f: f.fire_a_total_ac*0.40468564, 
        fire_cost=lambda f: f.fire_runtime,
        mu_over_nu=lambda f: 1/(8e2*f.fire_selprob)
        )
    return df_v3_n2e6


def gdf_pyromes(fpath='/Users/val/projects/SIG/datasets/Pyromes_CONUS_20200206.gpkg'):
    duckdb.install_extension('spatial')
    duckdb.load_extension('spatial')
    df = duckdb.sql(f"""
        SELECT 
            pyrome AS pyrome_id,
            NAME,
            ST_AsText(geom) AS geom_wkt
        FROM ST_Read("{fpath}", layer="Pyromes_CONUS_20200206") AS pyr
        """).df()
    gdf = gpd.GeoDataFrame(df.drop(columns=['geom_wkt']), geometry=gpd.GeoSeries.from_wkt(df.geom_wkt), crs=5070)
    return gdf


### Results

def assert_scalar_results():
    df_ff = read_df_ff_v3_n2e6()
    assert 1.5588234703893717 == is_optimal_gain(df_ff)
    assert 1.0903206908614824 ==  1/score_weight_fn(df_ff.assign(pyr_w=pyrome_weights_series(df_ff)), lambda f: f.pyr_w, [])
    df1pyr = df_ff[lambda f: (f.pyrome_id == 33)&(f.fire_tstop_hr > 0)]
    assert 1.3795962047686874 == is_optimal_gain(df1pyr)


### Figures

def plot_countours_effgain(df):
    eps = numpy.linspace(0.8, 2.0, 50)
    log10_s0 = numpy.linspace(-1.0, 2.3, 40)
    def effgain(eps, log10_s0):
        s0 = 10**log10_s0
        def fn_fire_cost(f):
            return 1 + (f.fire_size/s0)**eps
        return is_optimal_gain(df.assign(fire_cost=fn_fire_cost))
    x = eps
    y = log10_s0
    z = numpy.array([[effgain(e, ls0) for e in x] for ls0 in y])
    fig, ax = plt.subplots(dpi=600)
    CS = ax.contour(x, y, z, 
        levels=[1.01, 1.05, *numpy.arange(1.1, 1.8, 0.1), *numpy.arange(2.0, 2.4, 0.2), *numpy.arange(2.5, 5.0, 0.5)]
        #colors=['black']
        )
    ax.clabel(CS, inline=True, fontsize=10)
    ax.set_xlabel('$\\epsilon$')
    ax.set_ylabel('$A_0$ (ha)')
    ys = numpy.array([0.1, 0.5, 1, 5, 10, 50, 100])
    ax.set_yticks(numpy.log10(ys), labels=[str(y) for y in ys])
    return fig, ax


def fig1_countours():
    """
    Plots `fig:contours-gain-cost-size-model`.
    """
    plot_countours_effgain(read_df_hist_fires())


def gdf_pyrome_reweighting():
    df_ff = read_df_ff_v3_n2e6()
    dfpyr = df_pyrome_multipliers(df_ff)
    gdf_pyr = gdf_pyromes()
    gdf_pyr = gdf_pyr.merge(dfpyr, how='left', on='pyrome_id')
    return gdf_pyr


def table1_pyromes_reweighting():
    """
    Computes and prints some LaTeX for `tab:pyrome-multiplier`
    """
    gdf_pyr = gdf_pyrome_reweighting()
    n_sim_years = 2023-2011+1
    df_table = pandas.DataFrame([
            {"Pyrome id": r.pyrome_id,
            #"Pyrome name": r.NAME,
            "Frequency multiplier": '{:1.2f}'.format(r.pyr_mul),
            "Burned area (ha/yr)": '{:.1f}'.format(0.40468564*r.esum_burned_ac/n_sim_years),
            "Runtime (s/yr)": '{:.2f}'.format(r.esum_runtime_s/n_sim_years),
            "Fires (fire/yr)": '{:.1f}'.format(r.esum_1/n_sim_years)}
            for _i, r in gdf_pyr.iterrows()
        ])
    print(df_tab_pyrome_multipliers(gdf_pyr, n_sim_years).to_latex(index=False))


def fig2_map_pyrome_multipliers():
    """
    Plots `fig:map-pyrome-multipliers`.
    """
    gdf_pyr = gdf_pyrome_reweighting()
    fig, ax = plt.subplots(figsize=(12, 8), dpi=600)
    gdf_pyr.plot(
        legend=True, column='pyr_mul',
        label="Frequency multiplier",
        cmap=sns.color_palette("coolwarm", as_cmap=True),
        norm=matplotlib.colors.LogNorm(vmin=1/2.5, vmax=2.5),
        legend_kwds=dict(shrink=0.7),
        ax=ax)
    gdf_pyr.apply(lambda x: ax.annotate(text=x['pyrome_id'], xy=x.geometry.centroid.coords[0], ha='center', fontsize='x-small', alpha=0.5), axis=1)
    plt.tick_params(
        axis='both',
        which='both'
        #bottom=False, top=False, left=False, right=False,
        #labelbottom=False, labeltop=False, labelleft=False, labelright=False
        )
    ax.set_xlabel("x (m)")
    ax.set_ylabel("y (m)")
    return fig, ax



def plot_w_interpolation_tstop(df, w_fn, params):
    fig, ax = plt.subplots(dpi=600)
    w = w_fn(df, *params)
    norm_scalr = norm_scalar(df, w)
    w = norm_scalr * w
    w_optim = (df.fire_cost/df.fire_size)**0.5
    w_optim = w_optim * norm_scalar(df, w_optim)
    tstop_colname = "Fire duration (hr)"
    y_colname = "Frequency multiplier"
    dfo = df.assign(**{y_colname:1/w_optim, tstop_colname: lambda f: f.fire_tstop_hr}).sample(1000)
    sns.scatterplot(dfo,
        x=tstop_colname, 
        y=y_colname,
        marker='x',
        alpha=0.5,
        label="Optimal",
        ax=ax)
    dfl = df.assign(**{y_colname: 1/w, tstop_colname: lambda f: f.fire_tstop_hr}).sample(10000)
    sns.lineplot(dfl, x=tstop_colname, y=y_colname, label="Fitted", ax=ax)
    ax.set_xscale('log')


def fig3_1pyr_w1():
    """
    Plots `fig:w1-1pyr-frequency-multipliers`.
    """
    df_ff = read_df_ff_v3_n2e6()
    df1pyr = df_ff[lambda f: (f.pyrome_id == 33)&(f.fire_tstop_hr > 0)]
    w1_fit = fit_reweighting_model(df1pyr, w1, [0.0])
    plot_w_interpolation_tstop(df1pyr, w1, w1_fit['bestfit_params'])
    return w1_fit


def fig4_1pyr_w2():
    """
    Plots `fig:w2-1pyr-frequency-multipliers`.
    """
    df_ff = read_df_ff_v3_n2e6()
    df1pyr = df_ff[lambda f: (f.pyrome_id == 33)&(f.fire_tstop_hr > 0)]
    w2_fit = fit_reweighting_model(df1pyr, w2, [0.0, 5.0, 0.0, 150.0])
    plot_w_interpolation_tstop(df1pyr, w2, w2_fit['bestfit_params'])
    return w2_fit


def table2_w3_models():
    """
    Emits LaTeX for table `tab:w3-models-performance`.
    """
    df_ff = read_df_ff_v3_n2e6()
    modelA_kfold = w3_test_kfold(df_ff, feature_cols=w3_all_feature_cols, train_frac=0.5, n_iter=30)
    modelB_kfold = w3_test_kfold(df_ff, feature_cols=(w3_all_feature_cols - {'x', 'y'}), train_frac=0.5, n_iter=30)
    df_table = pandas.DataFrame([
        {
            "Model": model_name,            
            "Efficiency gain": '{:.2f} ($\sigma = {:.5f}$)'.format(numpy.mean(kf), numpy.std(kf, ddof=1))
        }
        for model_name, kf in [
            ('Model A', modelA_kfold),
            ('Model B', modelB_kfold)
        ]
    ])
    print(df_table.to_latex(index=False, escape=False, label='tab:w3-models-performance'))